/* PRESELECT module
 * by Alex Stieglecker <a.stieglecker@gmx.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 *
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

/*
 * Inclusions related to version problems. They *might* go after
 * <linux/module.h>, unless a symbol table is exported. All in all,
 * better putting them first anyways.
 */
#include <linux/config.h> /* retrieve the CONFIG_* macros */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#  define MODVERSIONS /* force it on */
#endif

#ifdef MODVERSIONS
#  include <linux/modversions.h>
#endif
/* end version stuff */
 
#include <linux/ip.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/config.h>
#include <linux/jhash.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/proc_fs.h>

#include <net/checksum.h>

#define PRESELECT_VERSION "0.0.1"


#define DEBUG_PRESELECT

#ifdef DEBUG_PRESELECT
#define DEBUGP	printk
#else
#define DEBUGP
#endif

#define NF_IP_PRI_SELECT -450 /* should be in netfilter_ipv4.h, NF_IP_PRI_ARPFAKE is less */

/* number of maximum ip adresses */
#define IPS_MAX 16

#define MAX_PARAM_LEN 80
#define MAX_NODE_NUM 16

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alexander Stieglecker <a.stieglecker@gmx.net>");
MODULE_DESCRIPTION("Packet-selection algorithm for loadbalancing");

static char *ip_list;               /* underscore seperated list of numeric ip addresses */
static uint16_t ips_num = 0;        /* number of ip adresses entered above */
static u_int32_t ips[IPS_MAX];      /* array of u_int32_t ip addresses entered above (host order) */

MODULE_PARM_DESC(ip_list, " underscore-seperated list of ip-addresses to ignore");
MODULE_PARM( ip_list, "s");

/* protects accesses to ips-array and ips_num */
DECLARE_RWLOCK(ips_lock);

static uint totalnodes = 1;            /* number of total nodes in cluster */

static uint localenode = 1;            /* index of local node in cluster 
                                        * just to hold the value of the load-parameter  */

static uint localenodes[MAX_NODE_NUM]; /* indices of nodes, locally administered */
static uint localenodes_num = 0;       /* number of locally administered nodes */

MODULE_PARM_DESC(totalnodes, " number of nodes in the cluster");
MODULE_PARM(totalnodes, "i");
MODULE_PARM_DESC(localenode, " number of nodes, locally administered");
MODULE_PARM(localenode, "i");

/* protects access to localenodes-array _and_ localenodes_num */
DECLARE_RWLOCK(localenodes_lock);

/* protects access to totalnodes */
DECLARE_RWLOCK(totalnodes_lock);

#ifdef CONFIG_PROC_FS
static struct file_operations node_proc_fops;
static struct file_operations ips_proc_fops;
static struct proc_dir_entry *preselect_procdir;
static struct proc_dir_entry *nodes_pde;         /* comma seperated list of
                                                  * locally administered nodes in the cluster */
static struct proc_dir_entry *totalnodes_pde;    /* number of total nodes in the cluster */
static struct proc_dir_entry *localips_pde;      /* comma seperated list of
                                                  * ips to ignore */
#endif

/** Splits the string 'str' into tokens. The array of characterarrays contains
  * the pointers to several tokens within the string. After splitting 'str' is
  * destroyed.
  * 'tokenarr' is an array of characterarrays (space must be provided by caller)
  * 'spitter' represents the seperator within the strings.
  * 'tokenarrlen' is the maximum number of tokens tokenarr can contain
  * 'tokennum' contains the actual number of tokens after execution 
  *
  * returns 1 if the splitting suceeded
  * returns 0 if the string was malformed.
  */
static int split_str(char *str, char splitter, int *tokennum, char **tokenarr, int maxtokens)
{
	int i;
	*tokennum=1;
	
	if(!str[0] || str[0]==splitter || maxtokens < 1) {
		*tokennum=0;
		return 0;
	}
	
	/* empty string means no error */
	if(str[0]=='\0') {
		*tokennum=0;
		return 1;
	}
	
	tokenarr[0] = str;
	for(i = 0; str[i] != '\0'; i++) {
		if(str[i] == splitter) {
			if(str[i+1]=='\0' || str[i+1]== splitter)
				return 0;
			str[i] = '\0';
			if(*tokennum >= maxtokens)
				return 0;
			tokenarr[(*tokennum)++] = str+i+1;
		}
	}
	
	return 1;
}

/* Converts the string representation of an ip into a uint32 number (host order)
 * After the conversion addr is destroyed.
 * Returns 1 if the conversion suceeded
 * Returns 0 if the string was malformed.
 */
int addr2ip(char *addr, u_int32_t *ip)
{
	int tokennum, tmp, i;
	u_int32_t ret=0;
	char *tokenarr[4], *invchar;

	if( !addr ) {
		return 0;
	}
	
	if(!split_str(addr, '.', &tokennum, tokenarr, 4)) {
		return 0;
	}
	
	if(tokennum != 4)
		return 0;

	for(i=0; i<4; i++) {
		ret = ret << 8;
		tmp = simple_strtoul(tokenarr[i], &invchar, 10);
		if(invchar[0]!='\0' || tmp > 255) return 0;
		ret = tmp | ret;
	}
	
	*ip = ret;
	return 1;
}

/* Parses the comma seperated list of numeric ips within the string ip_list.
 * The resulting ips are stored in 'ips', an array of integers in host order.
 * Returns 1 if the parsing process suceeded
 * Returns 0 if the string contains malformed ips.
 * Note: Parses a maximum number of IPS_MAX ip adresses.
 */
int parse_ips( char *ip_list, u_int32_t *ips, int *tokennum ) {

	u_int32_t ip;
	int i;
	char *tokenarr[IPS_MAX];

	if(*ip_list == '\0') {
		*tokennum = 0;
		return 1;
	}
	
	if(!split_str(ip_list, '_', tokennum, tokenarr, IPS_MAX)) {
		printk(KERN_WARNING "ip_preselect: invalid parameter\n");
		return 0;
	}
		
	for(i=0; i < *tokennum; i++) {
		if(!addr2ip(tokenarr[i], &ip)) {
			printk(KERN_WARNING "ip_preselect: invalid ip address\n");
			return 0;
		}

		ips[i] = ip;
	}

	return 1;
}

int add_node(unsigned int nodenum)
{
	int i;
	
	WRITE_LOCK(&localenodes_lock);
	READ_LOCK(&totalnodes_lock);
	
	if( localenodes_num >= totalnodes ) {
		WRITE_UNLOCK(&localenodes_lock);
		READ_UNLOCK(&totalnodes_lock);
		return -EINVAL;
	}
	
	if(nodenum > 0 && nodenum <= totalnodes) {
		/* is there already a node called 'nodenum'? */
		for(i=0; i < localenodes_num; i++) {
			if(localenodes[i] == nodenum) {
				WRITE_UNLOCK(&localenodes_lock);
				READ_UNLOCK(&totalnodes_lock);
				return -EINVAL;
			}
		}
		
		localenodes[localenodes_num] = nodenum;
		localenodes_num++;
		WRITE_UNLOCK(&localenodes_lock);
		READ_UNLOCK(&totalnodes_lock);
		return 0;
	}
	
	WRITE_UNLOCK(&localenodes_lock);
	READ_UNLOCK(&totalnodes_lock);
	
	return -EINVAL;
}

int del_node(unsigned int nodenum)
{
	int i;
	
	WRITE_LOCK(&localenodes_lock);

	if( localenodes_num == 0 ) {
		WRITE_UNLOCK(&localenodes_lock);
		return -EINVAL;
	}
	
	READ_LOCK(&totalnodes_lock);
	
	if(nodenum > 0 && nodenum <= totalnodes) {
		for(i=0; i < localenodes_num; i++) {
			if(localenodes[i] == nodenum) {
				memmove(&(localenodes[i]), &(localenodes[i+1]),
					 sizeof(uint)*(localenodes_num-(i+1)));
				localenodes_num--;
				READ_UNLOCK(&totalnodes_lock);
				WRITE_UNLOCK(&localenodes_lock);
				return 0;
			}
		}	
	}
	
	READ_UNLOCK(&totalnodes_lock);
	WRITE_UNLOCK(&localenodes_lock);

	return -EINVAL;
}

int add_ip(u_int32_t ip)
{
	int i;
	
	
	/* is there already this ip? */
	WRITE_LOCK(&ips_lock);
	
	if( ips_num >= IPS_MAX ) {
		WRITE_UNLOCK(&ips_lock);
		return -ENOMEM;
	}
	
	for(i=0; i < ips_num; i++) {
		if(ips[i] == ip) {
			WRITE_UNLOCK(&ips_lock);
			return -EINVAL;
		}
	}

	ips[ips_num] = ip;
	ips_num++;
	WRITE_UNLOCK(&ips_lock);
	return 0;

}

int del_ip(u_int32_t ip)
{
	int i;

	WRITE_LOCK(&ips_lock);
	
	if( ips_num == 0 ) {
		WRITE_UNLOCK(&ips_lock);
		return -EINVAL;
	}
	
	for(i=0; i < ips_num; i++) {
		if(ips[i] == ip) {
			memmove(&(ips[i]), &(ips[i+1]),
				 sizeof(u_int32_t)*(ips_num-(i+1)));
			ips_num--;
			WRITE_UNLOCK(&ips_lock);
			return 0;
		}
	}	
	WRITE_UNLOCK(&ips_lock);
	return -EINVAL;
}

int set_totalnodes(unsigned int newtotal)
{
	int i;
	
	READ_LOCK(&totalnodes_lock);
	if(newtotal == totalnodes) {
		READ_UNLOCK(&totalnodes_lock);
		return 1;
	}
	
	if(newtotal < localenodes_num) {
		READ_UNLOCK(&totalnodes_lock);
		return 0;
	}
	
	WRITE_LOCK(&totalnodes_lock);
	READ_UNLOCK(&totalnodes_lock);

	if(newtotal > totalnodes) {
		totalnodes = newtotal;
		WRITE_UNLOCK(&totalnodes_lock);
		return 1;
	}
	
	for(i=0; i < localenodes_num; i++) {
		if(newtotal < localenodes[i]) {
			WRITE_UNLOCK(&totalnodes_lock);
			return 0;
		}
	}
	
	totalnodes = newtotal;	
	WRITE_UNLOCK(&totalnodes_lock);
	return 1;
}

static inline int
is_responsible(u_int32_t nodenum, u_int32_t destip)
{
	int i;
	u_int32_t h_destip = ntohl(destip);

	READ_LOCK(&localenodes_lock);
	READ_LOCK(&ips_lock);
	if (localenodes_num == 0 && ips_num == 0) {
		READ_UNLOCK(&ips_lock);
		READ_UNLOCK(&localenodes_lock);
		return 0;
	}
	READ_UNLOCK(&ips_lock);

	for (i = 0; i < localenodes_num; i++) {
		if (localenodes[i] == nodenum) {
//			DEBUGP(KERN_DEBUG "ip_preselect: responsible because of hash\n");
			READ_UNLOCK(&localenodes_lock);
			return 1;
		}
	}
	

	READ_LOCK(&ips_lock);
	for (i = 0; i < ips_num; i++) {
		if (ips[i] == h_destip) {
			READ_UNLOCK(&ips_lock);
			READ_UNLOCK(&localenodes_lock);
//			DEBUGP(KERN_DEBUG "ip_preselect: responsible because of ip\n");
			return 1;
		}
	}

	READ_UNLOCK(&ips_lock);	
	READ_UNLOCK(&localenodes_lock);

	return 0;
}

/* Netfilter hook itself. */
unsigned int ip_preselect_in(unsigned int hooknum,
			     struct sk_buff **pskb,
			     const struct net_device *in,
			     const struct net_device *out,
			     int (*okfn)(struct sk_buff *))
{
	int hashval;
	struct iphdr *iph = (*pskb)->nh.iph;

	/* we need a commutative hash function 
	 * to ensure that all packets of a connection go over the same node */
	hashval = ntohl(iph->saddr) * ntohl(iph->daddr);
	hashval = jhash_1word(hashval, 0);
//	DEBUGP(KERN_DEBUG "ip_preselect: hashvalue: %u\n", (hashval % totalnodes)+1);
	
	READ_LOCK(&totalnodes_lock);
	if(is_responsible( (hashval % totalnodes)+1, iph->daddr)) {
//		DEBUGP(KERN_DEBUG "ip_preselect: responsible\n");
		READ_UNLOCK(&totalnodes_lock);
		return NF_ACCEPT;
	}
	READ_UNLOCK(&totalnodes_lock);

//	DEBUGP(KERN_DEBUG "ip_preselect: not responsible\n");
	return NF_DROP;

}

static struct nf_hook_ops ip_preselect_in_ops = {
	.hook		= ip_preselect_in,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0))
	.owner		= THIS_MODULE,
#endif
	.pf		= PF_INET,
	.hooknum	= NF_IP_PRE_ROUTING,
	.priority	= NF_IP_PRI_SELECT
};

/*********************************************************************** 
 * PROC DIR HANDLING 
 ***********************************************************************/

#ifdef CONFIG_PROC_FS
/* sequence file stuff */
static void *node_start(struct seq_file *m, loff_t *index)
{
	unsigned int *nodeidx;

	READ_LOCK(&localenodes_lock);
	if (*index >= localenodes_num)
		return NULL;

	nodeidx = kmalloc(sizeof(unsigned int), GFP_KERNEL);
	if (!nodeidx)
		return ERR_PTR(-ENOMEM);

	*nodeidx = *index;
	return nodeidx;
}


static void *node_next(struct seq_file *m, void *obj_ident, loff_t *index)
{
	unsigned int *nodeidx = (unsigned int *)obj_ident;

	*index = ++(*nodeidx);
	if (*index >= localenodes_num) {
		kfree(obj_ident);
		return NULL;
	}
	return nodeidx;
}

static void node_stop(struct seq_file *m, void *obj_ident)
{
	kfree(obj_ident);
	READ_UNLOCK(&localenodes_lock);
}

static int node_show(struct seq_file *s, void *obj_ident)
{
	unsigned int *nodeidx = (unsigned int *)obj_ident;
	
	if (*nodeidx != 0) 
		seq_putc(s, ',');
	seq_printf(s, "%u", localenodes[*nodeidx]);
	
	if (*nodeidx == localenodes_num-1)
		seq_putc(s, '\n');

	return 0;
}

static void *ips_start(struct seq_file *m, loff_t *index)
{
	unsigned int *ipsidx;

	READ_LOCK(&ips_lock);
	if (*index >= ips_num)
		return NULL;

	ipsidx = kmalloc(sizeof(unsigned int), GFP_KERNEL);
	if (!ipsidx)
		return ERR_PTR(-ENOMEM);

	*ipsidx = *index;
	return ipsidx;
}


static void *ips_next(struct seq_file *m, void *obj_ident, loff_t *index)
{
	unsigned int *ipsidx = (unsigned int *)obj_ident;

	*index = ++(*ipsidx);
	if (*index >= ips_num) {
		kfree(obj_ident);
		return NULL;
	}
	return ipsidx;
}

static void ips_stop(struct seq_file *m, void *obj_ident)
{
	kfree(obj_ident);
	READ_UNLOCK(&ips_lock);
}

static int ips_show(struct seq_file *s, void *obj_ident)
{
	unsigned int *ipsidx = (unsigned int *)obj_ident;
	
	if (*ipsidx != 0) 
		seq_putc(s, ',');
	seq_printf(s, "%u.%u.%u.%u", HIPQUAD(ips[*ipsidx]));
	
	if (*ipsidx == ips_num-1)
		seq_putc(s, '\n');

	return 0;
}

/* sequence file operations for node proc-entry */
static struct seq_operations node_sops = {
	.start = node_start,
	.next  = node_next,
	.stop  = node_stop,
	.show  = node_show,
};

/* sequence file operations for node proc-entry */
static struct seq_operations ips_sops = {
	.start = ips_start,
	.next  = ips_next,
	.stop  = ips_stop,
	.show  = ips_show,
};

/* end of sequence file stuff */

static int node_proc_open( struct inode *dev_node, struct file *f )
{
	return seq_open( f, &node_sops );
}

static ssize_t node_proc_write(struct file *file, const char *input,
				size_t size, loff_t *ofs)
{
#define PROC_WRITELEN	16
	char buffer[PROC_WRITELEN+1];
	unsigned int nodenum, i;

	if (copy_from_user(buffer, input, PROC_WRITELEN))
		return -EFAULT;
	
	for(i=0; i < PROC_WRITELEN; i++) {
		if( buffer[i]==10 ) break;
	}
	buffer[i] = '\0';
	
	if (*buffer == '+') {
		nodenum = simple_strtoul(buffer+1, NULL, 10);
		if (add_node(nodenum))
			return -ENOMEM;
	} else if (*buffer == '-') {
		nodenum = simple_strtoul(buffer+1, NULL,10);
		if (del_node(nodenum))
			return -ENOENT;
	} else
		return -EIO;

	return size;
}

static ssize_t ips_proc_write(struct file *file, const char *input,
				size_t size, loff_t *ofs)
{
	char buffer[PROC_WRITELEN+1];
	u_int32_t ip=0,i;

	if (copy_from_user(buffer, input, PROC_WRITELEN))
		return -EFAULT;
	
	for(i=0; i < PROC_WRITELEN; i++) {
		if( buffer[i]==10 ) break;
	}
	buffer[i] = '\0';
	
	DEBUGP(KERN_DEBUG "ip_preselect: buffer is %s\n", buffer);

	if (*buffer == '+') {
		if (!addr2ip(buffer+1, &ip)) {
			return -EINVAL;
		}
		if (add_ip(ip)) {
			DEBUGP(KERN_DEBUG "ip_preselect: adding error\n");
			return -ENOMEM;
		}
	} else if (*buffer == '-') {
		if (!addr2ip(buffer+1, &ip))
			return -EINVAL;
		if (del_ip(ip))
			return -ENOMEM;
	} else
		return -EIO;

	return size;
}

static int ips_proc_open( struct inode *dev_node, struct file *f )
{
	return seq_open( f, &ips_sops );
}

static int totalnodes_proc_read( char *buf, char **start, off_t offset,
                int size, int *eof, void *data)
{
	int ret;

	READ_LOCK(&totalnodes_lock);
	ret = snprintf(buf, size,"%d\n", totalnodes);
	READ_UNLOCK(&totalnodes_lock);

	*eof = 1;
	return ret;
}


static int totalnodes_proc_write( struct file *instanz, const char __user *userbuffer,
        unsigned long count, void *data )
{
	char kernel_buffer[PROC_WRITELEN+1];
	int newtotal, i;
	
	if (copy_from_user( kernel_buffer, userbuffer, PROC_WRITELEN ))
		return -EFAULT;
	
	for(i=0; i < PROC_WRITELEN; i++) {
		if( kernel_buffer[i]==10 ) break;
	}
	kernel_buffer[i] = '\0';
	
	DEBUGP(KERN_DEBUG "ip_preselect: buffer is %s\n", kernel_buffer);

	newtotal = simple_strtoul(kernel_buffer, NULL, 10);
	
	if( set_totalnodes(newtotal) )
		printk(KERN_ERR "ip_preselect: cannot change totalnodes, try to remove nodes\n");
	
	return count;
}

/* file operations for node proc-entry */
static struct file_operations node_proc_fops = {
	.owner	 = THIS_MODULE,
	.open	 = node_proc_open,
	.read	 = seq_read,
	.write   = node_proc_write,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

/* file operations for localips proc-entry */
static struct file_operations ips_proc_fops = {
	.owner	 = THIS_MODULE,
	.open	 = ips_proc_open,
	.read	 = seq_read,
	.write   = ips_proc_write,
	.llseek	 = seq_lseek,
	.release = seq_release,
};

#endif /* CONFIG_PROC_FS */

static int init_or_cleanup(int init)
{
	int ret = 0, tokennum, local_ips[IPS_MAX];
	
	if (!init) goto cleanup;

	/* parameter integrity checks */
	if(totalnodes < 2 || totalnodes > MAX_NODE_NUM) {
		printk(KERN_ERR "ip_preselect: parameter totalnode must be in between 2 and %d\n", MAX_NODE_NUM);
		ret = -EINVAL;
		goto cleanup_nothing;
	}
	
	if(localenode > totalnodes) {
		printk(KERN_ERR "ip_preselect: parameter localenode must not be greater than %d\n", totalnodes);
		ret = -EINVAL;
		goto cleanup_nothing;
	}
	
	if(localenode < 1) {
		printk(KERN_ERR "ip_preselect: parameter localenode must not be less than 1\n");
		ret = -EINVAL;
		goto cleanup_nothing;
	}

	memset(localenodes, 0, sizeof(localenodes));
	add_node(localenode);

	/* hook registration */
	ret = nf_register_hook(&ip_preselect_in_ops);
	if(ret < 0) {
		printk(KERN_WARNING "ip_preselect: can't register pre-routing selection hook.\n");
		goto cleanup_nothing;
	}

	if(!parse_ips(ip_list, local_ips, &tokennum)) {
		ret = -EINVAL;
		goto cleanup_hook;
	}

	WRITE_LOCK(&ips_lock);
	memcpy(ips, local_ips, sizeof(ips));
	ips_num = tokennum;
	WRITE_UNLOCK(&ips_lock);

#ifdef CONFIG_PROC_FS
	preselect_procdir = proc_mkdir("ip_preselect", proc_net);
	if (!preselect_procdir) {
		printk(KERN_ERR "ip_preselect: Unable to create proc dir entry 'ip_preselect'\n");
		ret = -ENOMEM;
		goto cleanup_root_procdir;
	}
	
	nodes_pde = create_proc_entry("nodes", S_IWUSR|S_IRUSR, preselect_procdir);
	if (!nodes_pde) {
		printk(KERN_ERR "ip_preselect: Unable to create proc dir entry 'nodes'\n");
		ret = -ENOMEM;
		goto cleanup_node_procdir;
	}
	nodes_pde->proc_fops = &node_proc_fops;
	
	totalnodes_pde = create_proc_entry("totalnodes", S_IWUSR|S_IRUSR, preselect_procdir);
	if (!totalnodes_pde) {
		printk(KERN_ERR "ip_preselect: Unable to create proc dir entry 'totalnodes'\n");
		ret = -ENOMEM;
		goto cleanup_totalnodes_procdir;
	}
	totalnodes_pde->read_proc  = totalnodes_proc_read;
	totalnodes_pde->write_proc = totalnodes_proc_write;
	
	localips_pde = create_proc_entry("localips", S_IWUSR|S_IRUSR, preselect_procdir);
	if (!localips_pde) {
		printk(KERN_ERR "ip_preselect: Unable to create proc dir entry\n");
		ret = -ENOMEM;
		goto cleanup_ips_procdir;
	}
	localips_pde->proc_fops = &ips_proc_fops;
#endif /* CONFIG_PROC_FS */

	printk(KERN_NOTICE "Preselect Version %s loaded successfully\n",
		PRESELECT_VERSION);

	return 0;

 cleanup: 

#ifdef CONFIG_PROC_FS
	remove_proc_entry(localips_pde->name, localips_pde->parent);
 cleanup_ips_procdir:
	remove_proc_entry(totalnodes_pde->name, totalnodes_pde->parent);
 cleanup_totalnodes_procdir:
	remove_proc_entry(nodes_pde->name, nodes_pde->parent);
 cleanup_node_procdir:
	remove_proc_entry(preselect_procdir->name, preselect_procdir->parent);
 cleanup_root_procdir:
#endif /* CONFIG_PROC_FS */
 
 cleanup_hook:
	nf_unregister_hook(&ip_preselect_in_ops);

 cleanup_nothing:
 	printk(KERN_NOTICE "Preselect Version %s unloaded\n",
		PRESELECT_VERSION);
	return ret;
}


static int __init init(void)
{
	return init_or_cleanup(1);
}

static void __exit fini(void)
{
	init_or_cleanup(0);
}

module_init(init);
module_exit(fini);
